from math import inf
import random
from typing import Tuple, List, Union
from dh2vrml.util import rand_color
import numpy as np
from x3d.x3d import (
    X3D,
    Scene, Viewpoint, NavigationInfo,
    Shape, Transform,
    Appearance, Material,
    Box, Cylinder, Extrusion,
    _X3DChildNode
)

from dh2vrml.dhparams import DhParams, JointType


UNIT_LENGTH = 1
# Keep track of last unit length for link extrusion calculations
PREV_UL = UNIT_LENGTH
UL = UNIT_LENGTH

def update_ul(new_ul: float):
    global UL, PREV_UL
    PREV_UL = UL
    UL = new_ul

def reset_ul():
    global UL
    UL = UNIT_LENGTH
    PREV_UL = UNIT_LENGTH

def revolute_joint(color : Tuple[float, float, float]) -> Transform:
    """Return a cylinder axial along the Z axis

    By default the x3d cylinder primitive is axial along the Y axis.
    By wrapping the primitive in a rotation 90 degrees about the X axis
    we get the desired orientation.
    """
    return Transform(
        rotation=(UL, 0, 0, np.pi/2),
        children=[
            Shape(
                geometry=Cylinder(
                    height=2*UL,
                    radius=0.5*UL
                ),
                appearance=Appearance(
                    material=Material(
                        diffuseColor=color
                    )
                )
            )
        ]
    )

def prismatic_joint(color : Tuple[float, float, float]) -> Shape: 
    return Shape(
        geometry=Box(
            size=(UL, UL, 2*UL)
        ),
        appearance=Appearance(
            material=Material(
                diffuseColor=color
            )
        )
    )

def link_cross_section() -> List[Tuple[float, float]]:
    radius = 0.1*UL
    segments = 30
    cross_section = []
    for i in range(segments):
        theta = (2*np.pi/segments)*i
        cross_section.append(
            (radius*np.cos(theta), radius*np.sin(theta),)
        )
    # Final point to close cross section profile
    cross_section.append((radius*np.cos(0), radius*np.sin(0),))
    return cross_section

def end_effector(color : Tuple[float, float, float]) -> Shape: 
    shaft_length = 1.5*UL
    gripper_width = 0.7*UL
    gripper_length = 0.7*UL
    shaft_spine = [(0, 0, 0), (0, 0, shaft_length)]
    left_spine = shaft_spine + [
        (-gripper_width/2, 0, shaft_length),
        (-gripper_width/2, 0, shaft_length + gripper_length),
    ]
    right_spine = shaft_spine + [
        (gripper_width/2, 0, shaft_length),
        (gripper_width/2, 0, shaft_length + gripper_length),
    ]

    return Transform(
        DEF=f'end_effector',
        children=[
            Shape(
                geometry=Extrusion(
                    crossSection=link_cross_section(),
                    spine=left_spine,
                    scale=[(1, 1)]*len(left_spine),
                    creaseAngle=100
                ),
                appearance=Appearance(
                    material=Material(
                        diffuseColor=color
                    )
                )
            ),
            Shape(
                geometry=Extrusion(
                    crossSection=link_cross_section(),
                    spine=right_spine,
                    scale=[(1, 1)]*len(right_spine),
                    creaseAngle=100
                ),
                appearance=Appearance(
                    material=Material(
                        diffuseColor=color
                    )
                )
            ),
        ]
    )

def get_link_body(idx: int, d: float, theta: float, r: float, alpha: float, color: Tuple[float, float, float]) -> Transform:
    def get_z(d, theta):
        return np.matrix([
            [np.cos(theta), -np.sin(theta), 0, 0],
            [np.sin(theta),  np.cos(theta), 0, 0],
            [0            ,  0            , 1, d],
            [0            ,  0            , 0, 1]
        ])

    def get_x(r, alpha):
        return np.matrix([
            [1, 0            ,  0            , r],
            [0, np.cos(alpha), -np.sin(alpha), 0],
            [0, np.sin(alpha),  np.cos(alpha), 0],
            [0, 0            ,  0            , 1]
        ])

    # Assume zero rotation in actuator, wrap geometry in a Transform frame with
    # rotation at the end for simplicity
    z = get_z(d, 0)
    x = get_x(r, alpha)
    t = z*x
    zero_point = np.matrix([0, 0, 0, 1]).transpose()
    final_point = t*zero_point
    # Extract x, y, z from final point
    final_point = tuple(final_point[0:3].transpose().tolist()[0])
    extrusion_spine = [
        (0, 0, 0),  # Start extrusion at center of joint
        (0, 0, 1.5*PREV_UL),  # Extrusion protrudes up in the Z axis out of joint
    ]
    if final_point[-1] <= 0:
        # Wrap link body around joint if next joint is lower
        extrusion_spine.extend([
            (PREV_UL, 0, 1.5*PREV_UL),
            (PREV_UL, 0, 0),
        ])
    extrusion_spine.append(final_point)

    return Transform(
        DEF=f'l{idx}_link_body',
        rotation=(0, 0, 1, theta),
        children=[
            Shape(
                geometry=Extrusion(
                    crossSection=link_cross_section(),
                    spine=extrusion_spine,
                    scale=[(1, 1)]*len(extrusion_spine),
                    creaseAngle=100
                ),
                appearance=Appearance(
                    material=Material(
                        diffuseColor=color
                    )
                )
            ),
        ]
    )

def get_joint(type: JointType, color: Tuple[float, float, float]) -> _X3DChildNode:
    if type == JointType.REVOLUTE:
        return revolute_joint(color)
    if type == JointType.PRISMATIC:
        return prismatic_joint(color)
    if type == JointType.END_EFFECTOR:
        # End effector should always be cyan
        return end_effector((0, 1, 1))

def get_link(idx: int, d: float, theta: float, r: float, alpha: float, joint_type: JointType, last_joint_type: JointType, color: Union[Tuple[float, float, float], None]=None) -> Tuple[Transform, Transform]:
    if not color:
        color = rand_color()
    joint = get_joint(joint_type, color)
    link_alpha = Transform(
        DEF=f'l{idx}_alpha',
        rotation=(1, 0, 0, alpha),
        children=[
            joint
        ]
    )
    link = Transform(
        DEF=f'l{idx}_{last_joint_type.name}',
        children=[
            get_link_body(idx, d, theta, r, alpha, color),
            Transform(
                DEF=f'l{idx}_d',
                translation=(0, 0, d),
                children=[
                    Transform(
                        DEF=f'l{idx}_theta',
                        rotation=(0, 0, 1, theta),
                        children=[
                            Transform(
                                DEF=f'l{idx}_r',
                                translation=(r, 0, 0),
                                children=[
                                    link_alpha
                                ]
                            )
                        ]
                    )
                ]
            )
        ]
    )
    return link, link_alpha


def base_model(base_joint : JointType) -> X3D:
    base_color = (0.2, 0.2, 0.2)
    return X3D(
        profile="Immersive", version="3.3",
        Scene=Scene(
            children=[
                NavigationInfo(
                    DEF="ExamineMode"
                ),
                Viewpoint(
                    orientation=(-1, 0, 0, 0.2),
                    position=(0, 5, 15),
                ),
                Transform(
                    DEF="Base",
                    translation=(0, 0, -1),
                    children=[
                        Shape(
                            geometry=Box(
                                size=(10, 10, 0.1)
                            ),
                            appearance=Appearance(
                                material=Material(
                                    diffuseColor=base_color
                                )
                            )
                        ),
                    ]
                ),
                get_joint(base_joint, base_color)
            ]
        )
    )


def build_x3d(params : DhParams) -> X3D:
    global UL
    parameters = params.params
    scale = params.scale
    colors = params.colors
    joint_types = params.joint_types
    base_joint = joint_types.pop(0)

    if scale[0] is not None:
        update_ul(scale[0])
    else:
        reset_ul()
    joint_types.append(JointType.END_EFFECTOR)
    model = base_model(base_joint)
    ptr = model.Scene

    last_joint_type = base_joint
    for idx, (p, j, c, s) in enumerate(zip(parameters, joint_types, colors, scale)):
        if s is not None:
            update_ul(s)
        else:
            update_ul(UL)
        # link 0 is the base "link"
        link_idx = idx + 1
        link, new_ptr = get_link(link_idx, p.d, p.theta, p.r, p.alpha, j, last_joint_type, c)
        ptr.children.append(link)
        ptr = new_ptr
        last_joint_type = j

    # Return UL to original length
    reset_ul()
    return model



import uuid
import sys
import tempfile
from typing import Tuple, Union, List
import os
import click

from dh2vrml.dhparams import DhParams
from dh2vrml.exporter import build_x3d

def get_params_from_stdin(file_ext: str) -> DhParams:
    with tempfile.TemporaryDirectory() as tmp_name:
        name = str(uuid.uuid1())
        file_path = os.path.join(tmp_name, f"{name}.{file_ext}")
        with open(file_path, "w") as f:
            f.writelines(sys.stdin)
        print(f"Wrote tempfile to {file_path}")
        return get_params_from_file(file_path)


def get_params_from_file(f: str) -> DhParams:
    print(f"Opening {f}")
    file_name = os.path.basename(f)
    name, ext = os.path.splitext(file_name)
    if ext.lower() == ".yaml":
        print("Parsing YAML file")
        params = DhParams.from_yaml(f)
    elif ext.lower() == ".csv":
        print("Parsing CSV file")
        params = DhParams.from_csv(f)
    elif ext.lower() == ".py":
        print("Importing Python file")
        params = DhParams.from_py(f)
    else:
        print(f"Error: unrecognized file extension {ext}")
        exit(1)
    return params

def write_x3d_file(file_name: str, params: DhParams) -> str:
    file_name = os.path.basename(file_name)
    name, ext = os.path.splitext(file_name)
    model = build_x3d(params)
    print("Checking XML serialization...")
    modelXML= model.XML()
    model.XMLvalidate()
    out_name = f"{name}.x3d"
    print(f"Writing output to {out_name}")
    with open(out_name, "w") as f:
        f.write(modelXML)
    return modelXML

@click.command()
@click.option(
    '-f', '--file', default=None, multiple=True,
    help='DH Parameter file (.yaml, .csv, .py)'
)
@click.option(
    '-s', '--stdin', default=None,
    help='Read input from stdin, specify filetype to parse as (yaml, csv, py)'
)
def main(
        file : Union[Tuple[str, ...], None],
        stdin : Union[str, None]
        ) -> List[str]:
    out = []
    if stdin:
        params = get_params_from_stdin(stdin)
        out.append(write_x3d_file(str(uuid.uuid1()), params))

    for f in file:
        params = get_params_from_file(f)
        out.append(write_x3d_file(f, params))

    return out


if __name__ == '__main__':
    main()



from importlib.machinery import SourceFileLoader
from enum import Enum
import pandas as pd
from typing import List, Dict, Union, Tuple
import yaml

class JointType(Enum):
    REVOLUTE = 1
    PRISMATIC = 2
    END_EFFECTOR = 3

class DhParam:
    def __init__(self, d, theta, r, alpha):
        self.d = d
        self.theta = theta
        self.r = r
        self.alpha = alpha

class DhParams:
    def __init__(self,
                 params: List[DhParam],
                 joint_types: List[JointType],
                 colors: Union[List[Union[Tuple[float, float, float], None]], None]=None,
                 scale: Union[List[Union[float, None]], None]=None):
        self.params = params
        self.joint_types = joint_types
        if colors:
            self.colors = colors
        else:
            self.colors = [None]*len(self.params)
        if scale:
            self.scale = scale
        else:
            self.scale = [1]*len(self.params)
        assert (
            len(self.params) == len(self.joint_types) == len(self.colors) == len(self.scale))

    @classmethod
    def from_yaml(cls, file_name: str):
        with open(file_name, "r") as f:
            dict_list = yaml.safe_load(f)
        return cls.from_dict_list(dict_list)

    @classmethod
    def from_csv(cls, file_name: str):
        df = pd.read_csv(file_name, sep="\s*[,]\s*", engine="python")
        df.fillna("", inplace=True)
        dict_list = df.to_dict("records")
        for d in dict_list:
            color = d.get("color")
            if color:
                color = color.strip()
                color = tuple([float(c) for c in color.split()])
                d["color"] = color
            else:
                d["color"] = None

            scale = d.get("scale")
            if not scale:
                d["scale"] = None

        return cls.from_dict_list(dict_list)

    @classmethod
    def from_py(cls, file_name: str):
        module = SourceFileLoader("m_dhparams", file_name).load_module()
        try:
            dict_list = module.params
        except AttributeError:
            print(f"Error: could not find `params` variable in {file_name}")
            exit(1)
        return cls.from_dict_list(dict_list)

    @classmethod
    def from_dict_list(cls, param_list : List[Dict[str, Union[str, int, float]]]):
        params = [
            DhParam(
                p["d"], p["theta"], p["r"], p["alpha"])
            for  p in param_list
        ]
        joints = [
            JointType[p["type"].upper()] if p.get("type") else JointType.REVOLUTE
            for p in param_list
        ]
        colors = [
            tuple(p["color"]) if p.get("color") else None
            for p in param_list
        ]
        scale = [p.get("scale") for p in param_list]
        return DhParams(params, joints, colors, scale)





# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dh2vrml']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'click>=8.0.3,<9.0.0',
 'numpy>=1.21,<2.0',
 'pandas>=1.3,<2.0',
 'x3d>=4.0.47,<5.0.0',
 'xmlschema>=1.9.2,<2.0.0']

entry_points = \
{'console_scripts': ['dh2vrml = dh2vrml.cli:main']}

setup_kwargs = {
    'name': 'dh2vrml',
    'version': '0.1.5',
    'description': 'Library and CLI tool to convert Denavit–Hartenberg parameters to an X3D model',
    'long_description': '# dh2vrml\n\ndh2vrml is a utility for converting Denavit–Hartenberg parameters into X3D models, with a particular focus on creating outputs suitable for use as a MATLAB Simulink VR Sink.\n\n## Installation\n\n```\npip install dh2vrml\n```\n\n## Usage\n\n```\ndh2vrml -f <file_name>\n```\n\n### Parameters\n\n- `type`: Joint type, either `revolute` or `prismatic`\n    - This refers to the joint at index `i - 1`, (i.e. the first joint is the base joint)\n- `d`, `theta`, `r`, `alpha`: DH parameters as specified on [Wikipedia](https://en.wikipedia.org/wiki/Denavit%E2%80%93Hartenberg_parameters)\n- `color`: Color of the joint and link at index `i`, in RGB format\n    - Values are floats ranging from 0 to 1\n    - The end effector is always colored `(0, 1, 1)`, (cyan)\n- `scale`: The relative size of joints and links\n    - Links are scaled cross sectionally (position is not affected)\n    - Joints are scaled volumetrically\n    - The first value scales both the base joint and the joint after it\n    - If no value is provided, the last provided value is used\n        - Scale of the model can be set globally by only providing `scale` for the first set of parameters\n\n### Supported file types\n\n#### YAML\n\n```yaml\n- type: revolute  \n  d: 1.5\n  theta: 0\n  r: 3\n  alpha: 0\n  color: [1, 0, 0]\n\n- type: revolute\n  d: 2\n  theta: 0\n  r: 4\n  alpha: 3.14159265359\n  color: [0, 0, 1]\n\n- type: prismatic\n  d: 3\n  theta: 0\n  r: 0\n  alpha: 0\n  color: [1, 0, 1]\n```\n\n#### CSV\n\n```csv\nd ,theta ,r ,alpha        ,type     , color\n2 ,0     ,0 ,1.5707963268 ,revolute , 1 0 0\n0 ,0     ,2 ,0            ,revolute , 0 1 0\n```\n\n#### Python\n\nFor the sake of making calculations involving `pi` easier, Python files are supported.\n\n> Beware: dh2vrml will blindly import and run whatever code is provided, always inspect the contents of the file before importing\n\n```py\nfrom math import pi\n\nparams = [\n    {\n        "type": "revolute",\n        "d": 2,\n        "theta": pi/2,\n        "r": 0,\n        "alpha": pi/2\n    },\n    {\n        "type": "revolute",\n        "d": 2,\n        "theta": pi/2,\n        "r": 0,\n        "alpha": -pi/2\n    },\n    {\n        "type": "revolute",\n        "d": 2,\n        "theta": pi/2,\n        "r": 2,\n        "alpha": 0\n    },\n]\n```\n',
    'author': 'Jasper Chan',
    'author_email': 'jasperchan515@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Gigahawk/dh2vrml',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
